package performancemonster.accumulatedcallgraph;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import performancemonster.accumulatedcallgraph.impl.MethodNodeImpl;
import performancemonster.accumulatedcallgraph.CallGraphChangeListener;
import performancemonster.stretchedcallgraph.Node;

/**
 * Analyzes elements and creates an accumulated call graph. This analyzer builds
 * on top of a {@link performancemonster.stretchedcallgraph.Analyzer}.
 */
public class Analyzer implements
    performancemonster.stretchedcallgraph.CallGraphChangeListener
{
  private final Set<CallGraphChangeListener> listeners;
  private final Map<Node, MethodNode> existingNodes;
  private final List<MethodNode> children;

  /**
   * Creates a new analyzer, building on top of the given
   * {@link performancemonster.stretchedcallgraph.Analyzer}.
   */
  public Analyzer(performancemonster.stretchedcallgraph.Analyzer analyzer)
  {
    existingNodes = new HashMap<Node, MethodNode>();
    listeners = new HashSet<CallGraphChangeListener>();
    children = new ArrayList<MethodNode>();
    analyzer.addChangeListener(this);
  }

  /**
   * Registers a {@link CallGraphChangeListener}.
   */
  public void addChangeListener(CallGraphChangeListener listener)
  {
    listeners.add(listener);
  }
  
  /**
   * Unregisters a {@link CallGraphChangeListener}.
   */
  public void removeChangeListener(CallGraphChangeListener listener)
  {
    listeners.remove(listener);
  }

  /**
   * Gives the children on the top level.
   */
  public List<MethodNode> getChildren()
  {
    return Collections.unmodifiableList(children);
  }

  public void threadAdded(performancemonster.stretchedcallgraph.ThreadNode node)
  {
    // Intentionally left blank
  }

  public void methodAdded(performancemonster.stretchedcallgraph.MethodNode node)
  {
    boolean creatingNewNode = false;
    MethodNode newNode = getSiblingWithSameName(node);
    if (newNode == null) {
      creatingNewNode = true;
      newNode = createNode(node);
    }
    newNode.addElement(node);
    existingNodes.put(node, newNode);
    for (CallGraphChangeListener listener : listeners) {
      if (creatingNewNode) {
        listener.methodAdded(newNode);
      } else {
        listener.methodUpdated(newNode);
      }
    }
    bubbleInPlace(newNode);
  }

  public void methodUpdated(
      performancemonster.stretchedcallgraph.MethodNode node)
  {
    MethodNode existingMethod = existingNodes.get(node);
    existingMethod.replaceElement(node);
    for (CallGraphChangeListener listener : listeners) {
      listener.methodUpdated(existingMethod);
    }
    bubbleInPlace(existingMethod);
  }

  public void errorAdded(
      performancemonster.stretchedcallgraph.CallGraphError error)
  {
    CallGraphError wrappedError = new performancemonster.accumulatedcallgraph.impl.CallGraphErrorImpl(
        error);
    for (CallGraphChangeListener listener : listeners) {
      listener.errorAdded(wrappedError);
    }
  }

  public void elementsCleared()
  {
    existingNodes.clear();
    children.clear();
    for (CallGraphChangeListener listener : listeners) {
      listener.elementsCleared();
    }
  }

  /**
   * Return a sibling of same name, if one exists. Returns null otherwise.
   */
  private MethodNode getSiblingWithSameName(
      performancemonster.stretchedcallgraph.MethodNode node)
  {
    List<MethodNode> siblings;
    performancemonster.stretchedcallgraph.Node parent = node.getParent();
    if (parent instanceof performancemonster.stretchedcallgraph.MethodNode) {
      MethodNode parentMethod = existingNodes.get(parent);
      siblings = parentMethod.getChildren();
    } else {
      siblings = children;
    }
    for (MethodNode child : siblings) {
      if (child.getName().equals(node.getName())) {
        return child;
      }
    }
    return null;
  }

  /**
   * Creates a new node, inserts it into the parent's list.
   */
  private MethodNode createNode(
      performancemonster.stretchedcallgraph.MethodNode node)
  {
    performancemonster.stretchedcallgraph.Node parent = node.getParent();
    if (parent instanceof performancemonster.stretchedcallgraph.MethodNode) {
      MethodNode parentMethod = existingNodes.get(parent);
      MethodNode newNode = new MethodNodeImpl(parentMethod);
      parentMethod.addChild(newNode);
      return newNode;
    } else {
      MethodNode newNode = new MethodNodeImpl(null);
      children.add(newNode);
      return newNode;
    }
  }

  /**
   * Puts the given node in the correct place among its siblings. Calls
   * {@link CallGraphChangeListener#methodsSwitched(MethodNode, MethodNode)}
   * on the listeners along the way.
   */
  private void bubbleInPlace(MethodNode node)
  {
    List<MethodNode> siblings;
    MethodNode parent = node.getParent();
    if (parent != null) {
      siblings = parent.getChildren();
    } else {
      siblings = children;
    }

    // Go through the sibling list, from the last node to the first, and order
    // the nodes. Since only one node has been updated, and since nodes only
    // gain their "total time", one run through the list is enough.
    for (int i = siblings.size() - 1; i > 0; i--) {
      MethodNode first = siblings.get(i - 1);
      MethodNode last = siblings.get(i);
      if (first.getTotalTime() < last.getTotalTime()) {
        siblings.set(i - 1, last);
        siblings.set(i, first);
        for (CallGraphChangeListener listener : listeners) {
          listener.methodsSwitched(first, last);
        }
      }
    }
  }
}
