// PerformanceMonster
// Copyright (C) 2005  Ole Friis <olefriis@gmail.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

package performancemonster.graph;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

public class GraphModel
{
  private static final Comparator<GraphElement> comparator = new GraphElementComparator();
  private final SortedSet<GraphElement> elements;
  private final Set<Long> elementsInFocus;
  private final Set<GraphModelFocusListener> focusListeners;
  private final Set<GraphModelChangeListener> changeListeners;
  
  public GraphModel()
  {
    elements = new TreeSet<GraphElement>(comparator);
    elementsInFocus = new HashSet<Long>();
    focusListeners = new HashSet<GraphModelFocusListener>();
    changeListeners = new HashSet<GraphModelChangeListener>();
  }
  
  public synchronized void clear()
  {
    elements.clear();
    for (GraphModelChangeListener listener : changeListeners) {
      listener.elementsCleared();
    }
  }
  
  public synchronized void putElement(GraphElement element)
  {
    elements.add(element);
    for (GraphModelChangeListener listener : changeListeners) {
      listener.elementAdded(element);
    }
  }
  
  public synchronized void replaceElement(GraphElement oldElement,
      GraphElement newElement)
  {
    elements.remove(oldElement);
    elements.add(newElement);
    for (GraphModelChangeListener listener : changeListeners) {
      listener.elementReplaced(oldElement, newElement);
    }
  }
    
  public synchronized Set<GraphElement> getElements()
  {
    return Collections.unmodifiableSet(elements);
  }
  
  public synchronized void addFocusListener(GraphModelFocusListener listener)
  {
    focusListeners.add(listener);
  }
  
  public synchronized void removeFocusListener(GraphModelFocusListener listener)
  {
    focusListeners.remove(listener);
  }

  public synchronized void addChangeListener(GraphModelChangeListener listener)
  {
    changeListeners.add(listener);
  }
  
  public synchronized void removeChangeListener(GraphModelChangeListener listener)
  {
    changeListeners.remove(listener);
  }
  
  /**
   * Specifies the IDs for the elements in focus. It is up to each view of the
   * GraphModel to interpret what it means for an element to be "in focus", e.g.
   * painting the elements in a specific color.
   */
  public synchronized void setElementsInFocus(Set<Long> elementsInFocus)
  {
    this.elementsInFocus.clear();
    this.elementsInFocus.addAll(elementsInFocus);
    for (GraphModelFocusListener listener : focusListeners) {
      listener.elementsInFocusChanged();
    }
  }
  
  /**
   * Gives a list of IDs for elements in focus. It is up to each view of the
   * GraphModel to interpret what it means for an element to be "in focus", e.g.
   * painting the elements in a specific color.
   */
  public synchronized Set<Long> getElementsInFocus()
  {
    return Collections.unmodifiableSet(elementsInFocus);
  }
  
  private static class GraphElementComparator implements Comparator<GraphElement>
  {
    public int compare(GraphElement e1, GraphElement e2)
    {
      long l = e1.getId() - e2.getId();
      if (l > 0) {
        return 1;
      } else if (l < 0) {
        return -1;
      }
      return 0;
    }
  }
}
