package performancemonster.presentation.mainwindow;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.File;
import java.util.Collections;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JToggleButton;
import javax.swing.KeyStroke;

import org.apache.log4j.Logger;

import performancemonster.external.GraphExporter;
import performancemonster.external.GraphImporter;
import performancemonster.external.exceptions.ExternalException;
import performancemonster.graph.GraphElement;
import performancemonster.graph.GraphModel;
import performancemonster.graph.GraphModelFocusListener;
import performancemonster.graph.coloring.Colorer;
import performancemonster.presentation.about.AboutDialog;
import performancemonster.presentation.accumulatedcallgraph.AccumulatedCallGraphPanel;
import performancemonster.presentation.colors.ColorGUI;
import performancemonster.presentation.graphelement.GraphElementFrame;
import performancemonster.presentation.helpers.Dialogs;
import performancemonster.presentation.helpers.LookAndFeelHelper;
import performancemonster.presentation.instrumentation.InstrumentGUI;
import performancemonster.presentation.overview.MonsterGUI;
import performancemonster.presentation.overview.impl.DescriptionListener;
import performancemonster.presentation.overview.impl.GraphArea;
import performancemonster.presentation.overview.impl.HoverListener;
import performancemonster.presentation.stretchedcallgraph.StretchedCallGraphPanel;

import com.vlsolutions.swing.docking.DockKey;
import com.vlsolutions.swing.docking.Dockable;
import com.vlsolutions.swing.docking.DockingConstants;
import com.vlsolutions.swing.docking.DockingDesktop;
import com.vlsolutions.swing.docking.DockingSelectorDialog;
import com.vlsolutions.swing.toolbars.ToolBarConstraints;
import com.vlsolutions.swing.toolbars.ToolBarContainer;
import com.vlsolutions.swing.toolbars.ToolBarPanel;
import com.vlsolutions.swing.toolbars.VLToolBar;

public class MainWindow
{
  private static final Logger logger = Logger.getLogger(MonsterGUI.class);
  
  private JToggleButton zoomMode;
  private JToggleButton updateMode;
  private GraphArea graph;
  private JLabel info;
  private JFrame frame;
  private File lastSavedFile;
  private JFileChooser fileChooser;
  private DockingDesktop desktop;
  private final GraphModel model;
  private final Colorer colorer;

  public MainWindow(GraphModel model)
  {
    this.model = model;
    this.colorer = new Colorer();
    model.addFocusListener(new GraphModelFocusListener() {
      public void elementsInFocusChanged()
      {
        if (graph != null) {
          graph.setElementsInFocus(MainWindow.this.model.getElementsInFocus());
        }
      }
    });
  }

  /**
   * Create the GUI and show it. For thread safety, this method should be
   * invoked from the event-dispatching thread.
   */
  public void createAndShowGUI()
  {
    LookAndFeelHelper.setCorrectLookAndFeel();

    // Create and set up the window.
    frame = new JFrame("Performance Monster");
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

    Action zoomOutAction = new AbstractAction("Zoom out")
    {
      public void actionPerformed(ActionEvent evt)
      {
        long oldBeginning = graph.getBeginning();
        long oldEnd = graph.getEnd();
        long diff = oldEnd - oldBeginning;
        graph.setBeginning(oldBeginning - diff / 2);
        graph.setEnd(oldEnd + diff / 2);
        graph.repaint();
      }
    };
    Action zoomModeAction = new AbstractAction("Zoom mode")
    {
      public void actionPerformed(ActionEvent evt)
      {
        graph.setZoomMode(zoomMode.isSelected());
      }
    };
    Action alignAction = new AbstractAction("Align")
    {
      public void actionPerformed(ActionEvent evt)
      {
        graph.align();
      }
    };
    Action updateAction = new AbstractAction("Update")
    {
      public void actionPerformed(ActionEvent evt)
      {
        // updateMode = updateMode.isSelected();
      }
    };
    Action clearAction = new AbstractAction("Clear")
    {
      public void actionPerformed(ActionEvent evt)
      {
        Set<GraphElement> empty = Collections.emptySet();
        model.clear();
        graph.setGraphElements(empty);
        graph.repaint();
      }
    };

    JButton zoomOut = new JButton(zoomOutAction);
    zoomMode = new JToggleButton(zoomModeAction);
    JButton reset = new JButton(alignAction);
    reset.setName("AlignButton");
    updateMode = new JToggleButton(updateAction);
    updateMode.setSelected(true);
    JButton clear = new JButton(clearAction);
    info = new JLabel("");
    graph = new GraphArea(colorer);

    graph.setBeginning(0L);
    graph.setEnd(300L);
    graph.addDescriptionListener(new DescriptionListener()
    {
      public void showDescription(GraphElement element)
      {
        new GraphElementFrame(element);
      }
    });
    graph.addHoverListener(new HoverListener()
    {
      public void mouseHovered(long position, GraphElement element)
      {
        info.setText("Position " + (position - graph.getBeginning()) + "ms: "
            + element);
      }
    });
    final AccumulatedCallGraphPanel callGraphGUI = new AccumulatedCallGraphPanel(model);
    final StretchedCallGraphPanel stretchedCallGraphGUI = new StretchedCallGraphPanel(model);

    Dockable dockableGraph = new Dockable()
    {
      private final DockKey key = new DockKey("Visual overview");

      public DockKey getDockKey()
      {
        return key;
      }

      public Component getComponent()
      {
        return graph;
      }
    };
    Dockable dockableCallGraph = new Dockable(){
      private final DockKey key = new DockKey("Accumulated call graph");

      public DockKey getDockKey()
      {
        return key;
      }

      public Component getComponent()
      {
        return callGraphGUI;
      }};
      Dockable dockableStretchedCallGraph = new Dockable(){
        private final DockKey key = new DockKey("Simple call graph");

        public DockKey getDockKey()
        {
          return key;
        }

        public Component getComponent()
        {
          return stretchedCallGraphGUI;
        }};

    desktop = new DockingDesktop();
    desktop.addDockable(dockableGraph);
    desktop
        .split(dockableGraph, dockableCallGraph, DockingConstants.SPLIT_LEFT);
    desktop.split(dockableCallGraph, dockableStretchedCallGraph,
        DockingConstants.SPLIT_BOTTOM);
    
    ToolBarContainer container = ToolBarContainer.createDefaultContainer(true,
        false, true, false);
    container.add(desktop, BorderLayout.CENTER);
    ToolBarPanel toolBarPanel = container.getToolBarPanelAt(BorderLayout.NORTH);
    VLToolBar toolBar = new VLToolBar();
    toolBar.add(zoomOut);
    toolBar.add(zoomMode);
    toolBar.add(reset);
    toolBar.add(updateMode);
    toolBar.add(clear);
    toolBar.add(info);
    toolBarPanel.add(toolBar, new ToolBarConstraints(0, 0));
    
    // Display the window.
    createMenus();
    frame.add(container);
    frame.pack();
    frame.setVisible(true);
  }
  
  public JFrame getJFrame()
  {
    return frame;
  }
  
  public void update()
  {
    Set<GraphElement> updatedList = model.getElements();
    graph.setGraphElements(updatedList);
    graph.setEnd(System.currentTimeMillis());
    graph.align();
    graph.setEnd(System.currentTimeMillis());
  }

  public boolean isFrozen()
  {
    if (updateMode != null) {
      return !updateMode.isSelected();
    } else {
      return true;
    }
  }

  private void createMenus()
  {
    JMenuBar menuBar = new JMenuBar();
    
    // File menu
    JMenu fileMenu = new JMenu("File");
    fileMenu.setMnemonic(KeyEvent.VK_F);
    menuBar.add(fileMenu);
    
    // View menu
    JMenu viewMenu = new JMenu("View");
    viewMenu.setMnemonic(KeyEvent.VK_V);
    menuBar.add(viewMenu);

    // Tools menu
    JMenu toolsMenu = new JMenu("Tools");
    viewMenu.setMnemonic(KeyEvent.VK_T);
    menuBar.add(toolsMenu);
    
    // Help menu
    JMenu helpMenu = new JMenu("Help");
    viewMenu.setMnemonic(KeyEvent.VK_H);
    menuBar.add(helpMenu);
    
    // File->Load Ctrl-L
    JMenuItem openItem = new JMenuItem("Open", KeyEvent.VK_O);
    openItem.setAccelerator(KeyStroke.getKeyStroke(
        KeyEvent.VK_O, ActionEvent.CTRL_MASK));
    openItem.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e)
      {
        openAction();
      }
    });
    fileMenu.add(openItem);

    // File->Save Ctrl-S
    JMenuItem saveItem = new JMenuItem("Save", KeyEvent.VK_S);
    saveItem.setAccelerator(KeyStroke.getKeyStroke(
        KeyEvent.VK_S, ActionEvent.CTRL_MASK));
    saveItem.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e)
      {
        saveAction();
      }
    });
    fileMenu.add(saveItem);
    
    // File->Save As...
    JMenuItem saveAsItem = new JMenuItem("Save as...", KeyEvent.VK_A);
    saveAsItem.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e)
      {
        saveAsAction();
      }
    });
    fileMenu.add(saveAsItem);

    fileMenu.addSeparator();
    
    // File->Quit Ctrl-Q
    JMenuItem quitItem = new JMenuItem("Quit");
    quitItem.setAccelerator(KeyStroke.getKeyStroke(
        KeyEvent.VK_Q, ActionEvent.CTRL_MASK));
    quitItem.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e)
      {
        quitAction();
      }
    });
    fileMenu.add(quitItem);
    
    // View->Manage docks
    JMenuItem manageDocksItem = new JMenuItem("Manage docks", KeyEvent.VK_D);
    manageDocksItem.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e)
      {
        DockingSelectorDialog dlg = new DockingSelectorDialog(frame);
        dlg.setDockingDesktop(desktop);
        dlg.setWizardLabelText("Choose which PerformanceMonster views to show and which to hide");
        dlg.pack();
        dlg.setLocationRelativeTo(frame);
        dlg.setVisible(true);
      }
    });
    viewMenu.add(manageDocksItem);
    
    // View->Colors
    JMenuItem colorsItem = new JMenuItem("Colors", KeyEvent.VK_C);
    colorsItem.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e)
      {
        new ColorGUI(colorer).show();
      }
    });
    viewMenu.add(colorsItem);

    // Tools->Instrument
    JMenuItem instrumentItem = new JMenuItem("Runtime instrumentation...",
        KeyEvent.VK_R);
    instrumentItem.addActionListener(new ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        InstrumentGUI.getInstance().show();
      }
    });
    toolsMenu.add(instrumentItem);

    // Help->About PerformanceMonster...
    JMenuItem aboutItem = new JMenuItem("About PerformanceMonster...",
        KeyEvent.VK_A);
    aboutItem.addActionListener(new ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        AboutDialog.showAboutDialog();
      }
    });
    helpMenu.add(aboutItem);

    frame.setJMenuBar(menuBar);
  }

  private void openAction()
  {
    if (getFileChooser().showOpenDialog(frame) == JFileChooser.APPROVE_OPTION) {
      File fileToOpen = getFileChooser().getSelectedFile();
      try {
        GraphImporter importer = new GraphImporter(model);
        importer.importFromFile(fileToOpen);
        lastSavedFile = fileToOpen;
        update();
        graph.align();
      } catch (ExternalException e) {
        String errorString = "Could not open file " + fileToOpen;
        logger.error(errorString, e);
        Dialogs.getInstance().showError(frame, errorString, e);
      }
    }
  }

  private void saveAction()
  {
    if (lastSavedFile == null) {
      saveAsAction();
    } else {
      saveToFile(lastSavedFile);
    }
  }
  
  private void saveAsAction()
  {
    if (getFileChooser().showSaveDialog(frame) == JFileChooser.APPROVE_OPTION) {
      File chosenFile = getFileChooser().getSelectedFile();
      if (chosenFile.exists()
          && !Dialogs.getInstance().showYesNo(
              frame,
              "The chosen file " + chosenFile.getName()
                  + " already exists.\nDo you want to overwrite?",
              "File already exists", "Yes", "No")) {
        return;
      }
      lastSavedFile = chosenFile;
      saveToFile(lastSavedFile);
    }
  }

  private void quitAction()
  {
    System.exit(0);
  }

  private void saveToFile(File file)
  {
    try {
      GraphExporter exporter = new GraphExporter(model);
      exporter.exportToFile(file);
    } catch (ExternalException e) {
      String errorString = "Could not save to file " + file;
      logger.error(errorString, e);
      Dialogs.getInstance().showError(frame, errorString, e);
    }
  }

  private JFileChooser getFileChooser()
  {
    if (fileChooser == null) {
      fileChooser = new JFileChooser();
    }
    return fileChooser;
  }
}
