package performancemonster.stretchedcallgraph;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import performancemonster.graph.GraphElement;
import performancemonster.graph.GraphModel;
import performancemonster.graph.GraphModelChangeListener;
import performancemonster.stretchedcallgraph.impl.CallGraphErrorImpl;

public class Analyzer implements GraphModelChangeListener
{
  private final Map<Long, ThreadNode> threads;
  private final Map<Long, MethodNode> elements;
  private final Set<CallGraphChangeListener> listeners;
  
  public Analyzer(GraphModel model)
  {
    threads = new HashMap<Long, ThreadNode>();
    elements = new HashMap<Long, MethodNode>();
    listeners = new HashSet<CallGraphChangeListener>();
    model.addChangeListener(this);

    // Feed the analyzed model with the graph model
    for (GraphElement element : model.getElements()) {
      elementAdded(element);
    }
  }
  
  public void addChangeListener(CallGraphChangeListener listener)
  {
    listeners.add(listener);
  }
  
  public void removeChangeListener(CallGraphChangeListener listener)
  {
    listeners.remove(listener);
  }
  
  public Collection<ThreadNode> getThreads()
  {
    return Collections.unmodifiableCollection(threads.values());
  }

  public void elementAdded(GraphElement element)
  {
    ThreadNode thread = getThreadNode(element);
    placeElement(element, thread, thread.getChildren());
  }
  
  private void placeElement(GraphElement element, Node parent,
      List<MethodNode> children)
  {
    if (children.isEmpty()) {
      putElement(element, parent);
    } else {
      MethodNode lastNode = children.get(children.size() - 1);
      if (!lastNode.isEnded() || lastNode.getEnd() >= element.getBeginning()) {
        placeElement(element, lastNode, lastNode.getChildren());
        if (lastNode.getEnd() < element.getEnd()) {
          addError("Child ends too late", "End of " + element.getMessage()
              + " is after end of " + lastNode.getName());
        }
      } else {
        putElement(element, parent);
      }
    }
  }
  
  private void putElement(GraphElement element, Node parent)
  {
    MethodNode node = (MethodNode) parent.addChild(element);
    elements.put(element.getId(), node);
    for (CallGraphChangeListener listener : listeners) {
      listener.methodAdded(node);
    }
    if (!element.isEnded() && parent.isEnded()) {
      addError("Unbounded child", "At least one child of " + parent.getName()
          + " is unbounded");
    }
  }
  
  private void addError(String category, String description)
  {
    CallGraphError error = new CallGraphErrorImpl(category, description);
    for (CallGraphChangeListener listener : listeners) {
      listener.errorAdded(error);
    }
  }
  
  public void elementReplaced(GraphElement oldElement, GraphElement newElement)
  {
    MethodNode node = elements.get(oldElement.getId());
    node.updateInformation(newElement);
    for (CallGraphChangeListener listener : listeners) {
      listener.methodUpdated(node);
    }
    
    if (!node.isEnded()) {
      return;
    }

    // See if we have introduced an inconsistency
    for (MethodNode child : node.getChildren()) {
      if (!child.isEnded()) {
        addError("Unbounded child", "At least one child of " + node.getName()
            + " is unbounded");
      } else if (child.getEnd() > node.getEnd()) {
        addError("Child ends too late", "At least one child of "
            + node.getName() + " ends too late");
      }
    }
  }

  public void elementsCleared()
  {
    threads.clear();
    for (CallGraphChangeListener listener : listeners) {
      listener.elementsCleared();
    }
  }

  private ThreadNode getThreadNode(GraphElement element)
  {
    if (!threads.containsKey(element.getThreadId())) {
      ThreadNode node = new ThreadNode(element.getThreadId());
      threads.put(element.getThreadId(), node);
      for (CallGraphChangeListener listener : listeners) {
        listener.threadAdded(node);
      }
    }
    return threads.get(element.getThreadId());
  }
}
