package performancemonster.test.accumulatedcallgraph;

import static org.easymock.EasyMock.*;

import java.util.List;
import java.util.Set;


import junit.framework.TestCase;
import performancemonster.accumulatedcallgraph.Analyzer;
import performancemonster.accumulatedcallgraph.CallGraphChangeListener;
import performancemonster.accumulatedcallgraph.CallGraphError;
import performancemonster.accumulatedcallgraph.impl.CallGraphErrorImpl;
import performancemonster.graph.GraphModel;
import performancemonster.stretchedcallgraph.MethodNode;
import performancemonster.stretchedcallgraph.ThreadNode;

public class AnalyzerTest extends TestCase
{
  private GraphModel model;
  private Analyzer analyzer;
  
  @Override
  protected void setUp() throws Exception
  {
    model = new GraphModel();
    performancemonster.stretchedcallgraph.Analyzer stretchedAnalyzer = new performancemonster.stretchedcallgraph.Analyzer(
        model);
    analyzer = new Analyzer(stretchedAnalyzer);
  }
  
  public void testEmptyAnalyzer()
  {
    assertEquals("Child list not empty", 0, analyzer.getChildren().size());
  }
  
  /**
   * When inserting a single element, the listener should get a methodAdded
   * call.
   */
  public void testInsertingSingleElement()
  {
    MethodNode node = createMethodNode(null, "DummyName", 1, 100, 1);
    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    listener.methodAdded(eqMethodNode(1));
    analyzer.addChangeListener(listener);

    replay(listener);
    analyzer.methodAdded(node);
    assertEquals("Wrong size of analyzer model", 1, analyzer.getChildren()
        .size());
    verify(listener);
  }

  /**
   * When updating a node, the listener should get a methodUpdated call.
   */
  public void testUpdatingElement()
  {
    MethodNode node = createMethodNode(null, "DummyName", 1, 100, 0);
    analyzer.methodAdded(node);
    reset(node);
    setExpectations(node, null, "DummyName", 1, 100, 150);
    
    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    listener.methodUpdated(eqMethodNode(1));
    analyzer.addChangeListener(listener);

    replay(listener);
    analyzer.methodUpdated(node);
    assertEquals("Wrong size of analyzer model", 1, analyzer.getChildren()
        .size());
    verify(listener);
  }

  /**
   * When inserting two methods with different name, the second insert should
   * cause a methodAdded call to the listener.
   */
  public void testInsertingTwoElementsWithDifferentName()
  {
    MethodNode node1 = createMethodNode(null, "DummyName", 1, 100, 1);
    analyzer.methodAdded(node1);
    MethodNode node2 = createMethodNode(null, "OtherDummyName", 2, 200, 1);

    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    listener.methodAdded(eqMethodNode(2));
    analyzer.addChangeListener(listener);

    replay(listener);
    analyzer.methodAdded(node2);
    assertEquals("Wrong size of analyzer model", 2, analyzer.getChildren()
        .size());
    verify(listener);
  }
  
  /**
   * When inserting two methods with same name, the second insert should cause
   * a methodUpdated call to the listener.
   */
  public void testInsertingTwoElementsWithSameNameInRoot()
  {
    MethodNode node1 = createMethodNode(null, "DummyName", 1, 100, 1);
    analyzer.methodAdded(node1);
    MethodNode node2 = createMethodNode(null, "DummyName", 2, 200, 1);

    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    listener.methodUpdated(eqMethodNode(1, 2));
    analyzer.addChangeListener(listener);

    replay(listener);
    analyzer.methodAdded(node2);
    assertEquals("Wrong size of analyzer model", 1, analyzer.getChildren()
        .size());
    verify(listener);
  }
  
  /**
   * When inserting two methods with same name, the second insert should cause
   * a methodUpdated call to the listener.
   */
  public void testInsertingTwoElementsWithSameNameOnSubnode()
  {
    MethodNode parent = createMethodNode(null, "Parent", 1, 100, 100);
    analyzer.methodAdded(parent);
    MethodNode node1 = createMethodNode(parent, "DummyName", 2, 100, 1);
    analyzer.methodAdded(node1);
    MethodNode node2 = createMethodNode(parent, "DummyName", 3, 200, 1);

    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    listener.methodUpdated(eqMethodNode(2, 3));
    analyzer.addChangeListener(listener);

    replay(listener);
    analyzer.methodAdded(node2);
    List<performancemonster.accumulatedcallgraph.MethodNode> roots = analyzer
        .getChildren();
    assertEquals("Wrong size of analyzer model", 1, roots.size());
    assertEquals("Wrong size of child list of single root node", 1, roots.get(0)
        .getChildren().size());
    verify(listener);
  }
  
  /**
   * When inserting an element that is bigger than an existing element, the new
   * element should be "bubbled in place".
   */
  public void testInsertingBiggerElement()
  {
    MethodNode node1 = createMethodNode(null, "DummyName", 1, 100, 1);
    analyzer.methodAdded(node1);
    MethodNode node2 = createMethodNode(null, "OtherDummyName", 2, 200, 100);

    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    listener.methodAdded(eqMethodNode(2));
    listener.methodsSwitched(eqMethodNode(1), eqMethodNode(2));
    analyzer.addChangeListener(listener);

    replay(listener);
    analyzer.methodAdded(node2);
    assertEquals("Wrong size of analyzer model", 2, analyzer.getChildren()
        .size());
    verify(listener);
  }
  
  /**
   * The element ids for an element should correspond to the ids for the
   * contained elements.
   */
  public void testGetElementIds()
  {
    setupSomeElements();
    List<performancemonster.accumulatedcallgraph.MethodNode> roots = analyzer
        .getChildren();

    Set<Long> dummyNameIds = roots.get(2).getElementIds();
    assertEquals("Wrong size of DummyName ids", 2, dummyNameIds.size());
    assertTrue("Required id not part of DummyName", dummyNameIds.contains(1L));
    assertTrue("Required id not part of DummyName", dummyNameIds.contains(2L));

    Set<Long> otherDummyNameIds = roots.get(1).getElementIds();
    assertEquals("Wrong size of OtherDummyName ids", 1, otherDummyNameIds
        .size());
    assertTrue("Required id not part of OtherDummyName", otherDummyNameIds
        .contains(3L));
  }
  
  /**
   * The total time for an element should correspond to the sum of the single
   * total times of the contained elements.
   */
  public void testGetTotalTime()
  {
    setupSomeElements();
    List<performancemonster.accumulatedcallgraph.MethodNode> roots = analyzer
        .getChildren();

    assertEquals("Wrong total time of ThirdDummyName", 300, roots.get(0)
        .getTotalTime());
    assertEquals("Wrong total time of OtherDummyName", 200, roots.get(1)
        .getTotalTime());
    assertEquals("Wrong total time of DummyName", 101, roots.get(2)
        .getTotalTime());
  }
  
  public void testIsEnded()
  {
    setupSomeElements();
    List<performancemonster.accumulatedcallgraph.MethodNode> roots = analyzer
        .getChildren();

    assertEquals("Wrong ending information for ThirdDummyName", false, roots
        .get(0).isEnded());
    assertEquals("Wrong ending information for OtherDummyName", true, roots
        .get(1).isEnded());
    assertEquals("Wrong ending information for DummyName", true, roots.get(2)
        .isEnded());
  }
  
  public void testGetNumberOfCalls()
  {
    setupSomeElements();
    List<performancemonster.accumulatedcallgraph.MethodNode> roots = analyzer
        .getChildren();

    assertEquals("Wrong number of calls for ThirdDummyName", 2, roots.get(0)
        .getNumberOfCalls());
    assertEquals("Wrong number of calls for OtherDummyName", 1, roots.get(1)
        .getNumberOfCalls());
    assertEquals("Wrong number of calls for DummyName", 2, roots.get(2)
        .getNumberOfCalls());
  }
  
  /**
   * Adding a new thread to the analyzer should have no effect whatsoever.
   */
  public void testAddingThread()
  {
    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    ThreadNode thread = new ThreadNode(1);

    replay(listener);
    analyzer.threadAdded(thread);
    assertEquals("Child list not empty", 0, analyzer.getChildren().size());
    verify(listener);
  }
  
  /**
   * When the listener has been removed, it should receive no more calls.
   */
  public void testRemovingListener()
  {
    MethodNode node = createMethodNode(null, "DummyName", 1, 100, 1);
    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    analyzer.addChangeListener(listener);
    analyzer.removeChangeListener(listener);

    replay(listener);
    analyzer.methodAdded(node);
    assertEquals("Wrong size of analyzer model", 1, analyzer.getChildren()
        .size());
    verify(listener);
  }
  
  public void testClearing()
  {
    MethodNode node1 = createMethodNode(null, "DummyName", 1, 100, 1);
    analyzer.methodAdded(node1);
    assertEquals("Wrong size of analyzer model", 1, analyzer.getChildren()
        .size());

    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    listener.elementsCleared();
    analyzer.addChangeListener(listener);

    replay(listener);
    analyzer.elementsCleared();
    assertEquals("Child list not empty", 0, analyzer.getChildren().size());
    verify(listener);
  }
  
  public void testError()
  {
    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    CallGraphError error = new CallGraphErrorImpl("Category", "Description");
    listener.errorAdded(error);
    analyzer.addChangeListener(listener);

    replay(listener);
    analyzer
        .errorAdded(new performancemonster.stretchedcallgraph.impl.CallGraphErrorImpl(
            "Category", "Description"));
    verify(listener);
  }
  
  private MethodNode createMethodNode(MethodNode parent, String name, long id,
      long endTime, long totalTime)
  {
    MethodNode node = createNiceMock(MethodNode.class);
    setExpectations(node, parent, name, id, endTime, totalTime);
    return node;
  }

  private void setExpectations(MethodNode node, MethodNode parent, String name,
      long id, long endTime, long totalTime)
  {
    expect(node.getParent()).andStubReturn(parent);
    expect(node.getName()).andStubReturn(name);
    expect(node.getElementId()).andStubReturn(id);
    expect(node.getEnd()).andStubReturn(endTime);
    expect(node.getTotalTime()).andStubReturn(totalTime);
    expect(node.isEnded()).andStubReturn(totalTime != 0);
    replay(node);
  }

  /**
   * After calling this method, the root of the analyzer model will have these
   * elements:
   * 
   * - ThirdDummyName: totalTime 300, ids 4 and 5, not ended.
   * - OtherDummyName: totalTime 200, id 3, ended.
   * - DummyName: totalTime 101, ids 1 and 2, ended.
   */
  private void setupSomeElements()
  {
    MethodNode node1 = createMethodNode(null, "DummyName", 1, 100, 1);
    MethodNode node2 = createMethodNode(null, "DummyName", 2, 200, 100);
    MethodNode node3 = createMethodNode(null, "OtherDummyName", 3, 400, 200);
    MethodNode node4 = createMethodNode(null, "ThirdDummyName", 4, 500, 300);
    MethodNode node5 = createMethodNode(null, "ThirdDummyName", 5, 200, 0);
    analyzer.methodAdded(node1);
    analyzer.methodAdded(node2);
    analyzer.methodAdded(node3);
    analyzer.methodAdded(node4);
    analyzer.methodAdded(node5);

    List<performancemonster.accumulatedcallgraph.MethodNode> roots = analyzer
        .getChildren();
    assertEquals("Wrong name of first element", "ThirdDummyName", roots.get(0)
        .getName());
    assertEquals("Wrong name of second element", "OtherDummyName", roots.get(1)
        .getName());
    assertEquals("Wrong name of third element", "DummyName", roots.get(2)
        .getName());
  }

  private static performancemonster.accumulatedcallgraph.MethodNode eqMethodNode(
      long... ids)
  {
    reportMatcher(new MethodNodeEquals(ids));
    return null;
  }
}
