package performancemonster.test.stretchedcallgraph;

import static org.easymock.EasyMock.*;

import java.util.Collection;
import java.util.List;

import junit.framework.TestCase;
import performancemonster.graph.GraphElement;
import performancemonster.graph.GraphModel;
import performancemonster.stretchedcallgraph.Analyzer;
import performancemonster.stretchedcallgraph.CallGraphChangeListener;
import performancemonster.stretchedcallgraph.MethodNode;
import performancemonster.stretchedcallgraph.ThreadNode;
import performancemonster.stretchedcallgraph.impl.CallGraphErrorImpl;
import performancemonster.stretchedcallgraph.impl.MethodNodeImpl;

public class AnalyzerTest extends TestCase
{
  private GraphModel model;
  private Analyzer analyzer;
  
  @Override
  protected void setUp() throws Exception
  {
    model = new GraphModel();
    analyzer = new Analyzer(model);
  }
  
  public void testEmptyAnalyzer()
  {
    assertEquals("Thread list not empty", 0, analyzer.getThreads().size());
  }
  
  public void testReplacingElement()
  {
    GraphElement openElement = createOpenGraphElement("Message", 100, 1);
    ThreadNode threadNode = new ThreadNode(0);
    MethodNode openNode = new MethodNodeImpl(threadNode, openElement);

    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    analyzer.addChangeListener(listener);
    listener.threadAdded(threadNode);
    listener.methodAdded(openNode);
    replay(listener);
    
    // Put an open element in the model
    model.putElement(openElement);
    verify(listener);

    Collection<ThreadNode> nodes = analyzer.getThreads();
    assertEquals("Incorret number of thread nodes", 1, nodes.size());
    threadNode = nodes.iterator().next();
    List<MethodNode> methodNodes = threadNode.getChildren();
    assertEquals("Incorrect number of method nodes", 1, methodNodes.size());
    MethodNode methodNode = methodNodes.get(0);
    assertEquals("Node should not be ended", false, methodNode.isEnded());
    try {
      methodNode.getTotalTime();
      fail("getTotalTime should fail when isEnded is false");
    } catch (IllegalStateException e) {
      // Intentionally left blank
    }
    try {
      methodNode.getTimeInMethod();
      fail("getTimeInMethod should fail when isEnded is false");
    } catch (IllegalStateException e) {
      // Intentionally left blank
    }
    assertEquals("Node should have no children", 0, methodNode.getChildren().size());
    assertEquals("Node id should match element id", openElement.getId(), methodNode.getElementId());
    assertEquals("Incorrect message", "Message", methodNode.getName());
    
    // End the element in the model
    GraphElement closedElement = createGraphElement("Message", 100, 200, 1);
    methodNode = new MethodNodeImpl(threadNode, closedElement);
    reset(listener);
    listener.methodUpdated(methodNode);
    replay(listener);
    model.replaceElement(openElement, closedElement);
    verify(listener);
    
    nodes = analyzer.getThreads();
    assertEquals("Incorret number of thread nodes", 1, nodes.size());
    threadNode = nodes.iterator().next();
    methodNodes = threadNode.getChildren();
    assertEquals("Incorrect number of method nodes", 1, methodNodes.size());
    methodNode = methodNodes.get(0);
    assertEquals("Node should be ended", true, methodNode.isEnded());
    assertEquals("Incorrect time in method", 100, methodNode.getTimeInMethod());
    assertEquals("Incorrect total time for method", 100, methodNode.getTotalTime());
    assertEquals("Node should have no children", 0, methodNode.getChildren().size());
    assertEquals("Node id should match element id", openElement.getId(), methodNode.getElementId());
    assertEquals("Incorrect message", "Message", methodNode.getName());
  }
  
  /**
   * When the end point of a child extends the end of the parent, an error
   * should be sent.
   */
  public void testTooLongChild()
  {
    // Put the parent in the model
    GraphElement parent = createGraphElement("Message", 100, 200, 1);
    model.putElement(parent);

    // Put a child, which is too long for the parent, into the model
    GraphElement child = createGraphElement("Message2", 150, 300, 2);
    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    analyzer.addChangeListener(listener);
    ThreadNode threadNode = new ThreadNode(0);
    MethodNode parentNode = new MethodNodeImpl(threadNode, parent);
    MethodNode childNode = new MethodNodeImpl(parentNode, child);
    CallGraphErrorImpl error = new CallGraphErrorImpl("Child ends too late",
        "End of Message2 is after end of Message");
    listener.methodAdded(childNode);
    listener.errorAdded(error);
    replay(listener);
    
    model.putElement(child);
    verify(listener);
  }

  /**
   * When a parent is ended, an error should be sent if a child is unbounded.
   */
  public void testEndingParentWithUnboundedChild()
  {
    // Put an open parent with a closed child in the model
    GraphElement parent = createOpenGraphElement("Message", 100, 1);
    GraphElement openChild = createOpenGraphElement("Message2", 150, 2);
    model.putElement(parent);
    model.putElement(openChild);
    
    // Close the parent, without closing the child
    GraphElement closedParent = createGraphElement("Message", 100, 200, 1);
    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    analyzer.addChangeListener(listener);
    ThreadNode threadNode = new ThreadNode(0);
    MethodNode parentNode = new MethodNodeImpl(threadNode, closedParent);
    CallGraphErrorImpl error = new CallGraphErrorImpl("Unbounded child",
    "At least one child of Message is unbounded");

    listener.methodUpdated(parentNode);
    listener.errorAdded(error);
    replay(listener);

    model.replaceElement(parent, closedParent);
    verify(listener);
  }

  /**
   * When a node is ended, it should not be possible to insert any more
   * unbounded children.
   */
  public void testNewUnboundedChildInBoundedParent()
  {
    // Put a closed node in the model
    GraphElement parent = createGraphElement("Message", 100, 200, 1);
    model.putElement(parent);
    
    // Put an open child into the parent
    GraphElement openChild = createOpenGraphElement("Message2", 150, 2);
    CallGraphChangeListener listener = createMock(CallGraphChangeListener.class);
    analyzer.addChangeListener(listener);
    ThreadNode threadNode = new ThreadNode(0);
    MethodNode parentNode = new MethodNodeImpl(threadNode, parent);
    MethodNode childNode = new MethodNodeImpl(parentNode, openChild);
    CallGraphErrorImpl error = new CallGraphErrorImpl("Unbounded child",
    "At least one child of Message is unbounded");

    listener.methodAdded(childNode);
    listener.errorAdded(error);
    replay(listener);

    model.putElement(openChild);
    verify(listener);
  }
  
  private GraphElement createOpenGraphElement(String message,
      long beginning, long id)
  {
    return createGraphElement(message, beginning, -1, id);
  }

  private GraphElement createGraphElement(final String message,
      final long beginning, final long end, final long id)
  {
    return new GraphElement()
    {
      public String getCategory()
      {
        return "Category";
      }
      
      public String getMessage()
      {
        return message;
      }
  
      public String getStackTrace()
      {
        return "Stack trace";
      }
  
      public boolean isEnded()
      {
        return end != -1;
      }

      public long getBeginning()
      {
        return beginning;
      }
  
      public long getEnd()
      {
        return end;
      }
      
      public long getThreadId()
      {
        return 0;
      }

      public long getId()
      {
        return id;
      }
    };
  }
}
